@@ -20,6 +20,15 @@ from ovsdbapp import constants as const
 from ovsdbapp import utils
 
 
+def normalize_prefix(addr):
+    return str(netaddr.IPNetwork(addr).cidr if '/' in addr
+               else netaddr.IPAddress(addr))
+
+
+def normalize_prefixes(addrs):
+    return [normalize_prefix(addr) for addr in addrs or []]
+
+
 class LsAddCommand(cmd.AddCommand):
     table_name = 'Logical_Switch'
 
@@ -181,13 +190,14 @@ class PgAclAddCommand(_AclAddHelper):
 
 class _AclDelHelper(cmd.BaseCommand):
     def __init__(self, api, entity, direction=None,
-                 priority=None, match=None):
+                 priority=None, match=None, if_exists=False):
         if (priority is None) != (match is None):
             raise TypeError("Must specify priority and match together")
         if priority is not None and not direction:
             raise TypeError("Cannot specify priority/match without direction")
         super().__init__(api)
         self.entity = entity
+        self.if_exists = if_exists
         self.conditions = []
         if direction:
             self.conditions.append(('direction', '=', direction))
@@ -197,7 +207,14 @@ class _AclDelHelper(cmd.BaseCommand):
                                     ('match', '=', match)]
 
     def run_idl(self, txn):
-        entity = self.api.lookup(self.lookup_table, self.entity)
+        try:
+            entity = self.api.lookup(self.lookup_table, self.entity)
+        except idlutils.RowNotFound as e:
+            if self.if_exists:
+                return
+            msg = "%s %s does not exist" % (self.lookup_table, self.entity)
+            raise RuntimeError(msg) from e
+
         for acl in [a for a in entity.acls
                     if idlutils.row_match(a, self.conditions)]:
             entity.delvalue('acls', acl)
@@ -207,12 +224,6 @@ class _AclDelHelper(cmd.BaseCommand):
 class AclDelCommand(_AclDelHelper):
     lookup_table = 'Logical_Switch'
 
-    def __init__(self, api, switch, direction=None,
-                 priority=None, match=None):
-        # NOTE: we're overriding the constructor here to not break any
-        # existing callers before we introduced Port Groups.
-        super().__init__(api, switch, direction, priority, match)
-
 
 class PgAclDelCommand(_AclDelHelper):
     lookup_table = 'Port_Group'
@@ -242,8 +253,7 @@ class AddressSetAddCommand(cmd.AddCommand):
     def __init__(self, api, name, addresses=None, may_exist=False):
         super().__init__(api)
         self.name = name
-        self.addresses = [str(netaddr.IPAddress(address))
-                          for address in addresses or []]
+        self.addresses = normalize_prefixes(addresses)
         self.may_exist = may_exist
 
     def run_idl(self, txn):
@@ -300,8 +310,7 @@ class AddressSetUpdateAddressesCommand(cmd.BaseCommand):
         self.address_set = address_set
         if isinstance(addresses, (str, bytes)):
             addresses = [addresses]
-        self.addresses = [str(netaddr.IPAddress(address))
-                          for address in addresses]
+        self.addresses = normalize_prefixes(addresses)
 
 
 class AddressSetAddAddressesCommand(AddressSetUpdateAddressesCommand):
@@ -853,7 +862,7 @@ class LrpAddCommand(cmd.BaseCommand):
         self.networks = [str(netaddr.IPNetwork(net)) for net in networks]
         self.router = router
         self.port = port
-        self.peer = peer
+        self.peer = peer if peer else []
         self.may_exist = may_exist
         self.columns = columns
         super().__init__(api)
@@ -1137,9 +1146,117 @@ class BFDGetCommand(cmd.BaseGetRowCommand):
     table = 'BFD'
 
 
+class MirrorAddCommand(cmd.AddCommand):
+    # cmd.AddCommand uses self.table_name, other base commands use self.table
+    table_name = 'Mirror'
+
+    def __init__(self, api, name, mirror_type, index, direction_filter, dest,
+                 external_ids=None, may_exist=False):
+        self.name = name
+        self.direction_filter = direction_filter
+        self.mirror_type = mirror_type
+        if mirror_type != 'local':
+            self.dest = str(netaddr.IPAddress(dest))
+        else:
+            self.dest = dest
+        self.index = index
+
+        super().__init__(api)
+
+        self.columns = {
+            'name': name,
+            'filter': direction_filter,
+            'sink': dest,
+            'type': mirror_type,
+            'index': index,
+            'external_ids': external_ids or {},
+        }
+        self.may_exist = may_exist
+
+    def run_idl(self, txn):
+        try:
+            mirror_result = self.api.lookup(self.table_name, self.name)
+            self.result = rowview.RowView(mirror_result)
+            if self.may_exist:
+                self.set_columns(mirror_result, **self.columns)
+                self.result = rowview.RowView(mirror_result)
+                return
+            raise RuntimeError("Mirror %s already exists" % self.name)
+        except idlutils.RowNotFound:
+            pass
+
+        mirror = txn.insert(self.api.tables[self.table_name])
+        mirror.name = self.name
+        mirror.type = self.mirror_type
+        mirror.filter = self.direction_filter
+        mirror.sink = self.dest
+        mirror.index = self.index
+        self.set_columns(mirror, **self.columns)
+        # Setting the result to something other than a :class:`rowview.RowView`
+        # or :class:`ovs.db.idl.Row` typed value will make the parent
+        # `post_commit` method retrieve the newly insterted row from IDL and
+        # return that to the caller.
+        self.result = mirror.uuid
+
+
+class MirrorDelCommand(cmd.DbDestroyCommand):
+    table = 'Mirror'
+
+    def __init__(self, api, record):
+        super().__init__(api, self.table, record)
+
+
+class MirrorGetCommand(cmd.BaseGetRowCommand):
+    table = 'Mirror'
+
+
+class LspAttachMirror(cmd.BaseCommand):
+    def __init__(self, api, port, mirror, may_exist=False):
+        super().__init__(api)
+        self.port = port
+        self.mirror = mirror
+        self.may_exist = may_exist
+
+    def run_idl(self, txn):
+        try:
+            lsp = self.api.lookup('Logical_Switch_Port', self.port)
+            mirror = self.api.lookup('Mirror', self.mirror)
+            if mirror in lsp.mirror_rules and not self.may_exist:
+                msg = "Mirror Rule %s is already set on LSP %s" % (self.mirror,
+                                                                   self.port)
+                raise RuntimeError(msg)
+            lsp.addvalue('mirror_rules', mirror)
+        except idlutils.RowNotFound as e:
+            raise RuntimeError("LSP %s not found" % self.port) from e
+
+
+class LspDetachMirror(cmd.BaseCommand):
+    def __init__(self, api, port, mirror, if_exist=False):
+        super().__init__(api)
+        self.port = port
+        self.mirror = mirror
+        self.if_exist = if_exist
+
+    def run_idl(self, txn):
+        try:
+            lsp = self.api.lookup('Logical_Switch_Port', self.port)
+            mirror = self.api.lookup('Mirror', self.mirror)
+            if mirror not in lsp.mirror_rules and not self.if_exist:
+                msg = "Mirror Rule %s doesn't exist on LSP %s" % (self.mirror,
+                                                                  self.port)
+                raise RuntimeError(msg)
+            lsp.delvalue('mirror_rules', mirror)
+        except idlutils.RowNotFound as e:
+            if self.if_exists:
+                return
+            msg = "LSP %s doesn't exist" % self.port
+            raise RuntimeError(msg) from e
+
+
 class LrRouteAddCommand(cmd.BaseCommand):
     def __init__(self, api, router, prefix, nexthop, port=None,
-                 policy='dst-ip', may_exist=False):
+                 policy='dst-ip', may_exist=False, ecmp=False,
+                 route_table=const.MAIN_ROUTE_TABLE, bfd=None):
         prefix = str(netaddr.IPNetwork(prefix))
         if nexthop != const.ROUTE_DISCARD:
             nexthop = str(netaddr.IPAddress(nexthop))
@@ -1149,12 +1266,21 @@ class LrRouteAddCommand(cmd.BaseCommand):
         self.nexthop = nexthop
         self.port = port
         self.policy = policy
+        self.ecmp = ecmp
+        self.route_table = route_table
+        self.bfd = bfd
         self.may_exist = may_exist
 
     def run_idl(self, txn):
         lr = self.api.lookup('Logical_Router', self.router)
         for route in lr.static_routes:
-            if self.prefix == route.ip_prefix:
+            if (
+                self.prefix == route.ip_prefix and
+                self.route_table == route.route_table and
+                "ic-learned-route" not in route.external_ids
+            ):
+                if self.ecmp and self.nexthop != route.nexthop:
+                    continue
                 if not self.may_exist:
                     msg = "Route %s already exists on router %s" % (
                         self.prefix, self.router)
@@ -1169,8 +1295,11 @@ class LrRouteAddCommand(cmd.BaseCommand):
         route.ip_prefix = self.prefix
         route.nexthop = self.nexthop
         route.policy = self.policy
+        route.route_table = self.route_table
         if self.port:
             route.output_port = self.port
+        if self.bfd:
+            route.bfd = self.bfd
         lr.addvalue('static_routes', route)
         self.result = route.uuid
 
@@ -1183,12 +1312,15 @@ class LrRouteAddCommand(cmd.BaseCommand):
 
 
 class LrRouteDelCommand(cmd.BaseCommand):
-    def __init__(self, api, router, prefix=None, if_exists=False):
+    def __init__(self, api, router, prefix=None, if_exists=False,
+                 nexthop=None, route_table=const.MAIN_ROUTE_TABLE):
         if prefix is not None:
             prefix = str(netaddr.IPNetwork(prefix))
         super().__init__(api)
         self.router = router
         self.prefix = prefix
+        self.nexthop = nexthop
+        self.route_table = route_table
         self.if_exists = if_exists
 
     def run_idl(self, txn):
@@ -1197,9 +1329,14 @@ class LrRouteDelCommand(cmd.BaseCommand):
             lr.static_routes = []
             return
         for route in lr.static_routes:
-            if self.prefix == route.ip_prefix:
+            if (
+                self.prefix == route.ip_prefix and
+                self.route_table == route.route_table
+            ):
+                if self.nexthop and route.nexthop != self.nexthop:
+                    continue
+
                 lr.delvalue('static_routes', route)
-                # There should only be one possible match
                 return
 
         if not self.if_exists:
@@ -1209,13 +1346,19 @@ class LrRouteDelCommand(cmd.BaseCommand):
 
 
 class LrRouteListCommand(cmd.ReadOnlyCommand):
-    def __init__(self, api, router):
+    def __init__(self, api, router, route_table=None):
         super().__init__(api)
         self.router = router
+        self.route_table = route_table
 
     def run_idl(self, txn):
         lr = self.api.lookup('Logical_Router', self.router)
-        self.result = [rowview.RowView(r) for r in lr.static_routes]
+        if self.route_table is not None:
+            self.result = [rowview.RowView(r)
+                           for r in lr.static_routes
+                           if r.route_table == self.route_table]
+        else:
+            self.result = [rowview.RowView(r) for r in lr.static_routes]
 
 
 class LrNatAddCommand(cmd.BaseCommand):
@@ -1532,15 +1675,22 @@ class LbDelHealthCheckCommand(cmd.BaseCommand):
             raise RuntimeError(msg)
 
 
-class LbAddIpPortMappingCommand(cmd.BaseCommand):
+class LbIpPortMappingCommand(cmd.BaseCommand):
+    @staticmethod
+    def normalize_ip(ip_str):
+        ip = netaddr.IPAddress(ip_str)
+        return f"[{ip}]" if ip.version == 6 else str(ip)
+
+
+class LbAddIpPortMappingCommand(LbIpPortMappingCommand):
     table = 'Load_Balancer'
 
     def __init__(self, api, lb, endpoint_ip, port_name, source_ip):
         super().__init__(api)
         self.lb = lb
-        self.endpoint_ip = str(netaddr.IPAddress(endpoint_ip))
+        self.endpoint_ip = self.normalize_ip(endpoint_ip)
         self.port_name = port_name
-        self.source_ip = str(netaddr.IPAddress(source_ip))
+        self.source_ip = self.normalize_ip(source_ip)
 
     def run_idl(self, txn):
         lb = self.api.lookup(self.table, self.lb)
@@ -1548,13 +1698,13 @@ class LbAddIpPortMappingCommand(cmd.BaseCommand):
                   '%s:%s' % (self.port_name, self.source_ip))
 
 
-class LbDelIpPortMappingCommand(cmd.BaseCommand):
+class LbDelIpPortMappingCommand(LbIpPortMappingCommand):
     table = 'Load_Balancer'
 
     def __init__(self, api, lb, endpoint_ip):
         super().__init__(api)
         self.lb = lb
-        self.endpoint_ip = str(netaddr.IPAddress(endpoint_ip))
+        self.endpoint_ip = self.normalize_ip(endpoint_ip)
 
     def run_idl(self, txn):
         lb = self.api.lookup(self.table, self.lb)
